"""Tests for xcube server source."""
import json
from unittest.mock import patch

import pytest
from requests.exceptions import ConnectionError as RequestsConnectionError
from requests.models import Response
from werkzeug.exceptions import ServiceUnavailable

from gateway_4d_viewer.interfaces import Tile
from gateway_4d_viewer.xcube_server_source import XCubeDataSource


@pytest.fixture()
def example_xcube_data_source() -> XCubeDataSource:
    return XCubeDataSource(server_url="http://please-give-data:8000/4d_viewer")


@pytest.fixture()
def example_data_set_meta_data() -> dict:
    return {
        "id": "local.c2rcc_flags",
        "last_modified": "2023-04-02T15:19:07.575709Z",
        "name": "Local OLCI L2C cube for region SNS C2RCC quality flags",
        "ui-type": "heatmap",
        "ui-path": "/user/local/c2rcc_flags"
    }


@pytest.fixture()
def example_data_set_config():
    return {
        "data-set": {
            "units": "",
            "value-range": {
                "min": 0.0,
                "max": 1.0
            },
            "times": [
                "2017-01-16T10:09:22Z",
                "2017-01-25T09:35:51Z",
                "2017-01-26T10:50:17Z",
                "2017-01-28T09:58:11Z",
                "2017-01-30T10:46:34Z"
            ],
            "ui-path": "/user/local/c2rcc_flags/",
            "ui-type": "heatmap",
            "last-modified": "2023-04-02T15:27:31.274803Z"
        },
        "tile-grid": {
            "extent": {
                "x": {
                    "min": 0,
                    "max": 5
                },
                "y": {
                    "min": 50,
                    "max": 52.5
                }
            },
            "origin": {
                "x": 0,
                "y": 52.5
            },
            "tile-size": {
                "x": 250,
                "y": 250
            },
            "lowest-resolution-tile-extent": {
                "x": 2.5,
                "y": 2.5
            },
            "tile-orientation": {
                "x": "positive",
                "y": "negative"
            },
            "number-of-resolutions": 3,
            "projection": "EPSG:4567",
        }
    }


def test__get_response_from_server(example_xcube_data_source: XCubeDataSource):
    with patch("gateway_4d_viewer.xcube_server_source.requests.get") as mock_get:

        mock_response = Response()
        mock_response.status_code = 200
        mock_response._content = b'All good'

        mock_get.return_value = mock_response
        response = example_xcube_data_source._get_data_from_server("dummy_endpoint", request_data={"give": "data"})
        assert response == "All good"
        mock_get.assert_called_once_with(url="http://please-give-data:8000/4d_viewer/dummy_endpoint",
                                         data=json.dumps({"give": "data"}), timeout=(5, 15))


def test__get_response_from_server_connection_error(example_xcube_data_source: XCubeDataSource):
    with patch("gateway_4d_viewer.xcube_server_source.requests.get") as mock_get:

        mock_response = Response()
        mock_response.status_code = 200
        mock_response._content = b'All good'

        mock_get.side_effect = RequestsConnectionError("Server unavailable")
        with pytest.raises(ServiceUnavailable, match="Server unavailable"):
            example_xcube_data_source._get_data_from_server("dummy_endpoint", request_data={"give": "data"})


def test_data_set_configuration(example_xcube_data_source: XCubeDataSource, example_data_set_config: dict):
    with patch.object(XCubeDataSource, '_get_data_from_server') as mock_get_data_from_server:

        mock_get_data_from_server.return_value = json.dumps(example_data_set_config)
        example_xcube_data_source.data_set_configuration(data_set_id="dummy_data_set_id")
        mock_get_data_from_server.assert_called_once_with(endpoint_string="dummy_data_set_id/configuration")


def test_data_sets(example_xcube_data_source: XCubeDataSource,
                   example_data_set_meta_data: dict):
    with patch.object(XCubeDataSource, '_get_data_from_server') as mock_get_data_from_server:

        mock_get_data_from_server.return_value = json.dumps({"variables": [example_data_set_meta_data]})

        example_xcube_data_source._data_sets()

        mock_get_data_from_server.assert_called_once_with(endpoint_string="variables")


def test_data_set_tiles(example_xcube_data_source: XCubeDataSource):
    with patch.object(XCubeDataSource, '_get_data_from_server') as mock_get_data_from_server:
        example_xcube_data_source.data_set_tiles(Tile("mock_dataset", "mock_resolution", "t", "x", "y", "z", "png"))
        mock_get_data_from_server.assert_called_once_with(
            endpoint_string="mock_dataset/tiles/mock_resolution/x_y_z_t_png",
            as_flask_response=True)



"""Unit tests interaces."""
from __future__ import annotations

import json
from unittest.mock import patch

import pytest
from schema import SchemaError

from gateway_4d_viewer.interfaces import DataSourceInterface
from gateway_4d_viewer.interfaces import Tile


class DummyDataSourceInterface(DataSourceInterface):
    """Dummy datasource to implement abstract class."""

    def _data_set_configuration(self: DummyDataSourceInterface, _: str) -> dict:
        return None

    def _data_sets(self: DummyDataSourceInterface) -> dict:
        return None

    # This abstract method isn't validated so we leave a stub here
    # The doc string D102 requirement is suppressed as unnecessary here.
    def data_set_tiles(self: DummyDataSourceInterface, _tile_definition: Tile) -> str:   # noqa:D102 - dummy method.
        return None

    def __eq__(self: DummyDataSourceInterface, _other_file: object) -> bool:   # noqa:D105 - dummy method.
        return True


def _fetch_json(file_path: str) -> list | dict:
    with open(file_path) as json_file:
        contents = json.load(json_file)
    return contents


def test_data_set_configuration():
    with patch.object(DummyDataSourceInterface, "_data_set_configuration") as mock_data_set_configuration:

        example_configuration = _fetch_json('tests/json_response_examples/configuration.json')
        mock_data_set_configuration.return_value = example_configuration

        interface = DummyDataSourceInterface()
        interface.data_set_configuration("test")


def test_data_set_configuration_2d():
    with patch.object(DummyDataSourceInterface, "_data_set_configuration") as mock_data_set_configuration:

        example_configuration = _fetch_json('tests/json_response_examples/configuration_2d.json')
        mock_data_set_configuration.return_value = example_configuration
        DummyDataSourceInterface().data_set_configuration("test")


def test_data_set_configuration_fail():
    interface = DummyDataSourceInterface()
    with pytest.raises(SchemaError):
        interface.data_set_configuration("test")


def test_data_sets():
    with patch.object(DummyDataSourceInterface, "_data_sets") as mock_data_set_configuration:

        example_data_sets = _fetch_json('tests/json_response_examples/data_sets.json')
        mock_data_set_configuration.return_value = example_data_sets

        DummyDataSourceInterface().data_sets()


def test_data_sets_fail():
    interface = DummyDataSourceInterface()
    with pytest.raises(SchemaError):
        interface.data_sets()



"""The module defines abstract interfaces and objects used throughout the middle tier."""
from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from typing import List

from flask import Response
from schema import Schema

from gateway_4d_viewer.api_schema import CONFIGURATION_SCHEMA
from gateway_4d_viewer.api_schema import DATA_SETS_SCHEMA


@dataclass
class Tile:
    """Encapsulates tile data specification (and simplifies function signatures)."""

    data_set_id: str
    resolution: str
    time: str
    x: str
    y: str
    z: str
    extension: str


class DataSourceInterface(ABC):
    """Abstract class definition for data source implementations.

    Note - obscure implementation specific exceptions should be caught and reraised with standard exception objects
    from werkzeug.exceptions.
    e.g. the following for missing resources:
    from werkzeug.exceptions import NotFound
    """

    def data_sets(self: DataSourceInterface) -> List[dict]:
        """
        Query available data sets for a given child data source implementation.

        Results are validated by the schema before being returned.

        Returns
        -------
        List[dict]
            Data sets from child implementation with added schema validation.
        """
        data_sets = self._data_sets()
        Schema(DATA_SETS_SCHEMA).validate(data_sets)
        return data_sets

    @ abstractmethod
    def _data_sets(self: DataSourceInterface) -> List[dict]:
        """Return data sets information. Child must implement this! See data_sets() for more detail."""
        raise NotImplementedError  # pragma: no cover

    def data_set_configuration(self: DataSourceInterface, data_set_id: str) -> dict:
        """
        Return data set configuration info.

        Results of child implementation are validated before being returned.

        Parameters
        ----------
        data_set_id : str
            Data set identifier.

        Returns
        -------
        dict
            Data set configuration dictionary.
        """
        configuration = self._data_set_configuration(data_set_id)
        Schema(CONFIGURATION_SCHEMA).validate(configuration)
        return configuration

    @ abstractmethod
    def _data_set_configuration(self: DataSourceInterface, _data_set_id: str) -> dict:
        """Return configuration/meta data for a given dataset. Child must implement this.

        See data_set_configuration for more detail.
        """
        raise NotImplementedError  # pragma: no cover

    @ abstractmethod
    def data_set_tiles(self: DataSourceInterface, _tile_definition: Tile) -> Response:
        """Fetch individual tiles."""
        raise NotImplementedError  # pragma: no cover

    @ abstractmethod
    def __eq__(self: DataSourceInterface, _other_object: object) -> bool:
        """
        Check object for equality.

        Two source objects should be equal if they point to the same source, e.g. the same google bucket or xcube
        server url.

        Parameters
        ----------
        _other_object : object
            The object to compare against this one.

        Returns
        -------
        bool
            Whether the two are equal
        """
        raise NotImplementedError  # pragma: no cover



"""Tests for google bucket source."""
import json
from unittest.mock import ANY
from unittest.mock import MagicMock
from unittest.mock import patch

from google.api_core import exceptions
from google.auth.exceptions import DefaultCredentialsError
from google.cloud.storage import Blob
from google.cloud.storage import Bucket
from google.cloud.storage import Client
import pytest
from werkzeug.exceptions import NotFound

from gateway_4d_viewer import app
from gateway_4d_viewer.google_bucket_source import _get_data_set_name_from_path
from gateway_4d_viewer.google_bucket_source import GoogleBucketDataSource
from gateway_4d_viewer.interfaces import Tile


@pytest.fixture()
def mock_client() -> MagicMock(Client):
    """
    Handle all client mocking in one place for the various use cases in tests.

    Returns
    -------
    MagicMock(Client)
        Mocked google Client object
    """
    blob_1 = MagicMock(spec=Blob)
    blob_1.name = "a/a.png"
    blob_1.public_url = "https://a/a.png"
    blob_2 = MagicMock(spec=Blob)
    blob_2.name = "a/b.png"
    blob_2.public_url = "https://a/b.png"
    blob_3 = MagicMock(spec=Blob)
    blob_3.name = "a/c/"
    blob_3.public_url = "shouldn't appear"

    # Blobs can be iterated for each blob or queried for prefixes
    mock_blobs = MagicMock()
    mock_blobs.__iter__.return_value = [blob_1, blob_2, blob_3]
    mock_blobs.prefixes = ['a/', 'b/']

    mock_client = MagicMock(spec=Client)
    mock_client.list_blobs.return_value = mock_blobs

    # Mock bucket calls
    mock_bucket = MagicMock(spec=Bucket)
    blob_1.download_to_filename.return_value = None
    blob_1.download_as_string.return_value = '{"A": "test"}'
    mock_bucket.blob.return_value = blob_1
    mock_client.get_bucket.return_value = mock_bucket

    return mock_client


@pytest.fixture()
def test_google_bucket_data_source(mock_client: MagicMock(Client)) -> GoogleBucketDataSource:
    return GoogleBucketDataSource("test_bucket", client=mock_client)


def test_create_instance_handles_credentials_error():
    with patch.object(Client, "__init__") as mock_client, \
            patch.object(Client, "create_anonymous_client") as mock_anonymous_client:
        mock_client.side_effect = DefaultCredentialsError('mocked error')
        GoogleBucketDataSource("test")
        # Expect attempt to create anonymous client instead
        mock_anonymous_client.assert_called_once()


def test_get_data_set_name_from_path():
    assert _get_data_set_name_from_path("/a/blah") == "blah"
    assert _get_data_set_name_from_path("/a/blah/") is None


def test_get_json_from_bucket(test_google_bucket_data_source: GoogleBucketDataSource):
    result = test_google_bucket_data_source._get_json_from_bucket("test/path")
    assert result == {'A': 'test'}


def test_send_file_from_bucket(test_google_bucket_data_source: GoogleBucketDataSource):
    with app.test_request_context() as _, \
            patch('gateway_4d_viewer.google_bucket_source.send_file') as mock_send_file:
        test_google_bucket_data_source.send_file_from_bucket("tmp.txt")
        mock_send_file.assert_called_once_with(ANY, download_name='tmp.txt')


def test_send_file_from_bucket_raises_not_found(test_google_bucket_data_source: GoogleBucketDataSource):

    # Reuse the existing mocked methods in the mock_client object but change the behaviour of download_to_filename
    mock_client = test_google_bucket_data_source.client
    mock_client.get_bucket().blob().download_to_filename.side_effect = exceptions.NotFound("")

    with app.test_request_context(), pytest.raises(NotFound):
        test_google_bucket_data_source.send_file_from_bucket("")


def test_data_set_configuration(test_google_bucket_data_source: GoogleBucketDataSource):
    with patch.object(GoogleBucketDataSource, '_get_json_from_bucket') as mock_get_json_from_bucket:
        test_google_bucket_data_source._data_set_configuration('data-set')
        mock_get_json_from_bucket.assert_called_once_with("api-v1/data-sets/data-set/configuration.json")


def test_data_sets(test_google_bucket_data_source: GoogleBucketDataSource):
    with patch.object(GoogleBucketDataSource, '_list_directories') as mock_paths, \
            patch.object(GoogleBucketDataSource, '_get_json_from_bucket') as mock_json:

        mock_paths.return_value = ['/a/b/c/']
        mock_json.return_value = {'data-set': {
            'ui-path': 'path',
            'ui-type': 'type',
            'last-modified': '1'
        }}

        data_sets = test_google_bucket_data_source._data_sets()

        assert data_sets[0]['id'] == 'c'
        assert data_sets[0]['last-modified'] == '1'
        assert data_sets[0]['ui-path'] == 'path'
        assert data_sets[0]['ui-type'] == 'type'


def test_data_sets_not_found_error(test_google_bucket_data_source: GoogleBucketDataSource):
    # Ensure google not found error is reraised as werkzeug NotFound
    with patch.object(GoogleBucketDataSource, '_list_directories') as mock_paths, \
            patch.object(GoogleBucketDataSource, '_get_json_from_bucket') as mock_json:

        mock_paths.return_value = ['/a/b/c/']
        mock_json.side_effect = exceptions.NotFound("")

        with pytest.raises(NotFound):
            test_google_bucket_data_source._data_sets()


def test_file_urls_in_directory(test_google_bucket_data_source: GoogleBucketDataSource):
    with app.test_request_context():
        results = test_google_bucket_data_source.file_urls_in_directory("A")     # Inputs here aren't used

    assert json.loads(results.get_data()) == [
        {'name': 'a'},
        {'name': 'b'}]


def test_list_blobs_with_prefix(test_google_bucket_data_source: GoogleBucketDataSource):
    results = test_google_bucket_data_source._list_directories("A")   # Inputs aren't used
    assert results == ['a/', 'b/']


def test_data_set_tiles(test_google_bucket_data_source: GoogleBucketDataSource):
    with patch.object(GoogleBucketDataSource, 'send_file_from_bucket') as mock_send_file:
        test_google_bucket_data_source.data_set_tiles(Tile("a", "r", "t", "x", "y", "z", "png"))
        mock_send_file.assert_called_once_with("api-v1/data-sets/a/tiles/r/x_y_z_t.png")



"""Schemas that define messages received via the API."""
from schema import Optional, Or, Regex


REGISTER_DATA_SOURCE_SCHEMA = Or({
    "data_source_type": "google_bucket_data_source",
    "bucket_name": str,
},
    {
    "data_source_type": "xcube_server_data_source",
    "server_url": str
})


# Iso date format "2023-03-08T01:39:51.352554Z" - sub second accuracy is optional.
ISO_DATE_SCHEMA = Regex(r'[0-9]{4}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}.*Z')
# Must start and end with a /
UI_PATH_SCHEMA = Regex(r'^/.*/$')
# Accepted layer types in the UI
UI_TYPE_SCHEMA = Or('flux',
                    'flux3d',
                    'heatmap',
                    'heatmap3d',
                    'heightmap',
                    'mesh')
# Used typically for floats that may or may not be persisted/cast as ints (e.g. resolution 500.0 vs 500)
NUMBER_SCHEMA = Or(float, int)

DATA_SET_SCHEMA = {
    # Unique identifier within the data source
    "id": str,
    # Datetime of last data set update
    "last-modified": ISO_DATE_SCHEMA,
    # Pretty name for the client
    "name": str,
    # Hierarchical path-like string to manage large number of datasets in the client.
    "ui-path": UI_PATH_SCHEMA,
    # Intended use by the client to determine what can be done with a given dataset.
    "ui-type": UI_TYPE_SCHEMA
}

DATA_SETS_SCHEMA = [DATA_SET_SCHEMA]

CONFIGURATION_SCHEMA = {
    "data-set": {
        # The units of the data values in the data set.
        "units": str,
        # This is the min/max value range of the data in the data set.
        # The main current use case is for denormalisation of data compressed into png/byte ranges.
        # e.g. a min: 5, max 10 would mean a png value byte would translate from 0->5 and 255->10.
        "value-range": {
            "min": float,
            "max": float
        },
        # A list of datetimes for which the data exists. (This tends to not be a regular axis)
        "times": [ISO_DATE_SCHEMA],
        # Hierarchical path-like string to manage large number of datasets in the client.
        "ui-path": UI_PATH_SCHEMA,
        # Intended use by the client to determine what can be done with a given dataset.
        "ui-type": UI_TYPE_SCHEMA,
        # Datetime of last data set update
        "last-modified": ISO_DATE_SCHEMA
    },
    "tile-grid": {
        # The extent of the dataset. Note: not necessarily a whole number of tiles!
        "extent": {
            "x": {
                "min": NUMBER_SCHEMA,
                "max": NUMBER_SCHEMA
            },
            "y": {
                "min": NUMBER_SCHEMA,
                "max": NUMBER_SCHEMA
            },
            Optional("z"): {
                "min": NUMBER_SCHEMA,
                "max": NUMBER_SCHEMA
            }
        },
        # The coordinates of the corner of the first tile in the tile grid. Whether that's the top/bottom, left/right is
        # determined by tile-orientation,
        "origin": {
            "x": NUMBER_SCHEMA,
            "y": NUMBER_SCHEMA,
            Optional("z"): NUMBER_SCHEMA
        },
        # The number of 'pixels' in each tile
        "tile-size": {
            "x": int,
            "y": int,
            Optional("z"): int
        },
        # The real-world size of a given tile at the lowest resolution. Other resolutions are deduced from this.
        "lowest-resolution-tile-extent": {
            "x": NUMBER_SCHEMA,
            "y": NUMBER_SCHEMA,
            Optional("z"): NUMBER_SCHEMA
        },
        # How the tile grid and the contents of each tile are oriented. By default, we try and follow the convention
        # that pngs are y-affine negative and x-affine positive. z is prefered as positive, too.
        "tile-orientation": {
            "x": Or("positive", "negative"),
            "y": Or("positive", "negative"),
            Optional("z"): Or("positive", "negative")
        },
        # How many resolution levels of data exist.
        "number-of-resolutions": int,
        # Which projection the tiles are served in. Must be in EPSG: or blank for now.
        Optional("projection"): Regex(r"^$|EPSG:[0-9]{4,5}$")
    }
}



"""Google bucket implementation of the abstract data source interface."""
from __future__ import annotations

import json
from typing import Dict, List

from flask import Response
import requests
from werkzeug.exceptions import ServiceUnavailable

from gateway_4d_viewer.interfaces import DataSourceInterface
from gateway_4d_viewer.interfaces import Tile


class XCubeDataSource(DataSourceInterface):
    """Implements the data source interface accessing data from an xcube server."""

    def __init__(self: XCubeDataSource, server_url: str) -> None:
        """
        Instantiate object.

        Parameters
        ----------
        server_url : str
            URL of the Xcube server to query for data.
        """
        super().__init__()
        self.server_url = server_url

    def _get_data_from_server(self: XCubeDataSource, endpoint_string: str, request_data: Dict = None,
                              as_flask_response: bool = False) -> str | Response:
        try:
            request_url = f"{self.server_url}/{endpoint_string}"
            request_data = json.dumps(request_data) if request_data is not None else None
            response = requests.get(url=request_url, data=request_data, timeout=(5, 15))
            assert response.ok
        except (requests.exceptions.ConnectionError, AssertionError, requests.exceptions.ReadTimeout) as exc:
            print(f"Failed to reach xcube server at {self.server_url}")
            raise ServiceUnavailable(exc)

        if as_flask_response:
            # translate from requests.Response to flask.Response - flask will throw an error otherwise
            return Response(response=response.content,
                            status=response.status_code,
                            headers=dict(response.headers))
        else:
            return response.text

    def _data_sets(self: XCubeDataSource) -> List[dict]:
        """
        Return a list of datasets.

        Returns
        -------
        List[dict]
            A list of datasets with basic meta data. See the abstract implementation for schema reference.
        """
        # protection for production - if xcube server unreachable, just exclude it from sources passed to the 4D client
        # without raising an error
        try:
            # xcube server doesn't allow endpoints to return a list, so we wrap it in a dict and unwrap here
            return json.loads(self._get_data_from_server(endpoint_string="variables"))["variables"]
        except ServiceUnavailable:
            return []

    def _data_set_configuration(self: XCubeDataSource, data_set_id: str) -> dict:
        """
        Return data set meta data for a specific data set id.

        Parameters
        ----------
        data_set_id : str
            Unique identifier of the data set

        Returns
        -------
        dict
            A meta data dictionary.
        """
        return json.loads(self._get_data_from_server(endpoint_string=f"{data_set_id}/configuration"))

    def data_set_tiles(self: XCubeDataSource, tile_definition: Tile) -> Response:
        """
        Get single tile of data.

        For now, if z is not set, it will be set to a string "None" in the server endpoint path. This is temporary;
        need to figure out if it is possible to assign multiple routes to a route handler within xcube. If not, we might
        end up with two separate endpoints instead - a 2d and a 3d one. TODO: update once sorted.

        Parameters
        ----------
        tile_definition : Tile
            Specification of the tile to retrieve.

        Returns
        -------
        Response
            Http response with object.
        """
        tile_path_template = \
            "{data_set_id}/tiles/{resolution}/{x}_{y}_{z}_{date}_{extension}"

        tile_path = tile_path_template.format(data_set_id=tile_definition.data_set_id,
                                              resolution=tile_definition.resolution,
                                              x=tile_definition.x,
                                              y=tile_definition.y,
                                              z=tile_definition.z if tile_definition.z is not None else "None",
                                              date=tile_definition.time,
                                              extension=tile_definition.extension
                                              )

        return self._get_data_from_server(endpoint_string=tile_path, as_flask_response=True)

    def __eq__(self: XCubeDataSource, other_object: object) -> bool:
        """
        Check object for equality.

        We consider two XCubeDataSource objects equal if they point to the same server url.

        Parameters
        ----------
        other_object : object
            The object to compare against this one.

        Returns
        -------
        bool
            Whether the two are equal.
        """
        if isinstance(other_object, XCubeDataSource):
            return other_object.server_url == self.server_url
        return False



"""Unit tests for gateway_4d_viewer.routes_v1.py."""
from http import HTTPStatus
import json
from unittest.mock import Mock
from unittest.mock import patch

from flask.testing import FlaskClient
import pytest
from werkzeug.exceptions import InternalServerError, NotFound

from gateway_4d_viewer import app
from gateway_4d_viewer import routes_v1
from gateway_4d_viewer.google_bucket_source import GoogleBucketDataSource
from gateway_4d_viewer.interfaces import DataSourceInterface
from gateway_4d_viewer.interfaces import Tile
from gateway_4d_viewer.xcube_server_source import XCubeDataSource


@pytest.fixture()
def client() -> FlaskClient:
    return app.test_client()


@pytest.fixture()
def file_urls_in_directory_example():
    return json.dumps([
        {'name': 'a', 'url': 'url_a'},
        {'name': 'b', 'url': 'url_b'}])


@pytest.fixture()
def mock_data_source():
    # Simple mock data source instance to use throughout tests.
    mock_source = Mock(spec=DataSourceInterface)
    mock_source.data_sets.return_value = [{'id': 'test_dataset'}]
    mock_source.data_set_configuration.return_value = {"att": "value"}
    return mock_source


@pytest.fixture()
def mock_data_sources(mock_data_source: Mock):
    # Mocks the DATA_SOURCES dictionary when required.
    with patch.dict('gateway_4d_viewer.routes_v1.DATA_SOURCES', {'source': mock_data_source}, clear=True):
        yield None


@pytest.fixture()
def mock_split_api_id(mock_data_source: Mock):
    # For endpoints that derive data set/sourve from the url, this mocks out values returned from _split_api_id
    with patch('gateway_4d_viewer.routes_v1._split_api_id') as mock_split_api_id:
        data_set = "A"
        mock_split_api_id.return_value = mock_data_source, data_set
        yield None


def test_api_catches_internal_error(client: FlaskClient):
    # use a simple end point to mimic an internal error.
    with patch.object(GoogleBucketDataSource, 'send_file_from_bucket') as mock_send_file:
        mock_send_file.side_effect = InternalServerError("Howdy")
        response = client.get("/api-v1/color-maps/id")
        assert response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR
        assert "Howdy" in response.text


@pytest.mark.usefixtures("mock_data_sources")
def test_get_data_source(mock_data_source: Mock):

    data_source, data_set = routes_v1._split_api_id(f"source{routes_v1.DATA_SOURCE_DELIMITER}set")
    assert data_source == mock_data_source
    assert data_set == "set"


@pytest.mark.usefixtures("mock_data_sources")
def test_get_data_sets(client: FlaskClient):

    response = client.get("/api-v1/data-sets")
    expected_api_id = f'source{routes_v1.DATA_SOURCE_DELIMITER}test_dataset'
    assert json.loads(response.get_data())[0]['id'] == expected_api_id


@pytest.mark.usefixtures("mock_data_sources")
def test_get_data_sets_catches_notfound(client: FlaskClient, mock_data_source: Mock):
    # Override data_sets side effect for this test only
    mock_data_source.data_sets.side_effect = NotFound("")
    response = client.get("/api-v1/data-sets")
    assert response.status_code == HTTPStatus.NOT_FOUND


@pytest.mark.usefixtures("mock_split_api_id")
def test_get_data_set_configuration(client: FlaskClient, mock_data_source: Mock):

    response = client.get("/api-v1/data-sets/any_api_key/configuration")
    # Check datasource method called
    mock_data_source.data_set_configuration.assert_called_once_with("A")
    # Check response is jsonified
    assert json.loads(response.get_data()) == {"att": "value"}


@pytest.mark.usefixtures("mock_split_api_id")
def test_get_data_set_configuration_catches_notfound(client: FlaskClient, mock_data_source: Mock):
    mock_data_source.data_set_configuration.side_effect = NotFound("")
    response = client.get("/api-v1/data-sets/any_api_key/configuration")
    assert response.status_code == HTTPStatus.NOT_FOUND


@pytest.mark.usefixtures("mock_split_api_id")
def test_return_tile(client: FlaskClient, mock_data_source: Mock):

    client.get("/api-v1/data-sets/any_api_key/tiles/resolution_time_x_y_png")

    # Check the mock data source is called correctly
    mock_data_source.data_set_tiles.assert_called_once_with(
        Tile(data_set_id="A", resolution="resolution", time="time", x="x", y="y", z=None, extension="png"))


@pytest.mark.usefixtures("mock_split_api_id")
def test_return_tile_with_z(client: FlaskClient, mock_data_source: Mock):

    client.get("/api-v1/data-sets/dataset_id/tiles/resolution_time_x_y_z_ewv")

    # Check the mock data source is called correctly
    mock_data_source.data_set_tiles.assert_called_once_with(
        Tile(data_set_id="A", resolution="resolution", time="time", x="x", y="y", z="z", extension="ewv"))


@pytest.mark.usefixtures("mock_split_api_id")
def test_return_tile_catches_notfound(client: FlaskClient, mock_data_source: Mock):

    mock_data_source.data_set_tiles.side_effect = NotFound("")
    response = client.get("/api-v1/data-sets/any_api_key/tiles/resolution_time_x_y_png")
    assert response.status_code == HTTPStatus.NOT_FOUND


def test_color_maps(client: FlaskClient, file_urls_in_directory_example: dict):

    with patch.object(GoogleBucketDataSource, 'file_urls_in_directory') as file_urls_in_directory:
        file_urls_in_directory.return_value = file_urls_in_directory_example
        response = client.get("/api-v1/color-maps")
        response_data = json.loads(response.get_data())
        assert response_data[0]['name'] == 'a'
        assert response_data[1]['url'] == 'url_b'


def test_color_map(client: FlaskClient):
    with patch.object(GoogleBucketDataSource, 'send_file_from_bucket') as mock_send_file:
        client.get("/api-v1/color-maps/id")
        mock_send_file.assert_called_once_with("api-v1/colour-maps/id.png")


def test_color_map_catches_notfound(client: FlaskClient):
    with patch.object(GoogleBucketDataSource, 'send_file_from_bucket') as mock_send_file:
        mock_send_file.side_effect = NotFound("")
        response = client.get("/api-v1/color-maps/id")
        assert response.status_code == HTTPStatus.NOT_FOUND


def test_configurations(client: FlaskClient, file_urls_in_directory_example: dict):

    with patch.object(GoogleBucketDataSource, 'file_urls_in_directory') as file_urls_in_directory:
        file_urls_in_directory.return_value = file_urls_in_directory_example
        response = client.get("/api-v1/configurations")
        response_data = json.loads(response.get_data())
        assert response_data[0]['name'] == 'a'
        assert response_data[1]['url'] == 'url_b'


def test_configuration(client: FlaskClient):
    with patch.object(GoogleBucketDataSource, 'send_file_from_bucket') as mock_send_file:
        client.get("/api-v1/configurations/id")
        mock_send_file.assert_called_once_with("api-v1/configurations/id.json")


def test_configuration_catches_notfound(client: FlaskClient):
    with patch.object(GoogleBucketDataSource, 'send_file_from_bucket') as mock_send_file:
        mock_send_file.side_effect = NotFound()
        response = client.get("/api-v1/configurations/id")
        assert response.status_code == HTTPStatus.NOT_FOUND


def test_register_xcube_data_source(client: FlaskClient, mock_data_source: Mock):
    with patch.dict('gateway_4d_viewer.routes_v1.DATA_SOURCES', {'source': mock_data_source}, clear=True)\
            as mock_data_sources:
        resp = client.post("/register-data-source/mock_xcube_source",
                           data=json.dumps({"data_source_type": "xcube_server_data_source",
                                            "server_url": "http://dummy-url:5000"}))
        assert resp.status_code == 200
        assert resp.text.endswith("mock_xcube_source added.")
        assert "mock_xcube_source" in mock_data_sources
        added_xcube_source = mock_data_sources["mock_xcube_source"]
        assert isinstance(added_xcube_source, XCubeDataSource)
        assert added_xcube_source.server_url == "http://dummy-url:5000"


def test_register_bucket_data_source(client: FlaskClient, mock_data_source: Mock):
    with patch.dict('gateway_4d_viewer.routes_v1.DATA_SOURCES', {'source': mock_data_source}, clear=True)\
            as mock_data_sources:
        resp = client.post("/register-data-source/mock_bucket_source",
                           data=json.dumps({"data_source_type": "google_bucket_data_source",
                                            "bucket_name": "mock_bucket_name"}))
        assert resp.status_code == 200
        assert resp.text.endswith("mock_bucket_source added.")
        assert "mock_bucket_source" in mock_data_sources
        added_bucket_source = mock_data_sources["mock_bucket_source"]
        assert isinstance(added_bucket_source, GoogleBucketDataSource)
        assert added_bucket_source.bucket_name == "mock_bucket_name"


def test_register_invalid_data_source(client: FlaskClient, mock_data_source: Mock):
    with patch.dict('gateway_4d_viewer.routes_v1.DATA_SOURCES', {'source': mock_data_source}, clear=True):
        resp = client.post("/register-data-source/mock_bucket_source",
                           data=json.dumps({"data_source_type": "invalid_source_type",
                                            "bucket_name": "mock_bucket_name"}))
        assert resp.status_code == HTTPStatus.BAD_REQUEST
        assert "Format invalid" in resp.text


@pytest.mark.usefixtures("mock_data_sources")
def test_register_data_source_error(client: FlaskClient):
    resp = client.post("/register-data-source/source",
                       data=json.dumps({"data_source_type": "google_bucket_data_source",
                                        "bucket_name": "mock_bucket_name"}))
    assert resp.status_code == 400
    assert resp.text.endswith("source already present.")


def test_deregister_data_source(client: FlaskClient, mock_data_source: Mock):
    with patch.dict('gateway_4d_viewer.routes_v1.DATA_SOURCES', {'source': mock_data_source}, clear=True)\
            as mock_data_sources:
        resp = client.delete("/deregister-data-source/source")
        assert resp.status_code == 200
        assert resp.text.endswith("source deleted.")
        assert "source" not in mock_data_sources


@pytest.mark.usefixtures("mock_data_sources")
def test_deregister_data_source_error(client: FlaskClient):
    resp = client.delete("/deregister-data-source/non-existent-source")
    assert resp.status_code == 400
    assert resp.text.endswith("non-existent-source not present, cannot delete.")



"""
Wrapper functions for dataset information.

This file and package will be deleted once v0 of the API is deprecated.
"""
import json

from google.cloud.storage import Client

from gateway_4d_viewer.google_bucket_source import GoogleBucketDataSource


def get_datasets(client: Client, bucket: str, path: str) -> str:
    """Build a list of available datasets and return as json.

    Parameters
    ----------
    client : Client
        A google cloud storage client instance
    bucket : str
        Google bucket
    path : str
        Google blob path prefix

    Returns
    -------
    str
        Json string.
    """
    delimiter = '/'
    bucket_data_source = GoogleBucketDataSource(bucket, client=client)
    sub_folders = bucket_data_source._list_directories(root_directory=path)
    clean_sub_folders = [sub_folder.replace(path, '', 1).replace(delimiter, '', 1) for sub_folder in sub_folders]
    clean_sub_folders.sort()

    dataset_items = []

    for sub_folder in clean_sub_folders:
        dataset = {
            'id': sub_folder,
            'title': sub_folder
        }

        dataset_items.append(dataset)

    datasets = {'datasets': dataset_items}

    return json.dumps(datasets)



"""Google bucket implementation of the abstract data source interface."""
from __future__ import annotations

import json
import logging
from pathlib import Path
import tempfile
from typing import List, Optional

from flask import jsonify
from flask import Response
from flask import send_file
from google.api_core.exceptions import NotFound as GoogleNotFound
from google.auth.exceptions import DefaultCredentialsError
from google.cloud.storage import Client
from werkzeug.exceptions import NotFound

from gateway_4d_viewer.interfaces import DataSourceInterface
from gateway_4d_viewer.interfaces import Tile

logger = logging.getLogger(__name__)


def _get_data_set_name_from_path(name: str) -> str:
    # Case ending in slash is parent directory
    if name[-1] == '/':
        return None
    return Path(name).stem


class GoogleBucketDataSource(DataSourceInterface):
    """Implements the data source interface accessing data from a propietary google bucket."""

    data_set_root_path = "api-v1/data-sets/"
    configuration_template = data_set_root_path + "{data_set_id}/configuration.json"

    def __init__(self: GoogleBucketDataSource, bucket_name: str, client: Optional[Client] = None) -> None:
        """
        Instantiate object.

        Parameters
        ----------
        bucket_name : str
            Name of the google bucket to retrieve data from.
        client : Client, optional
            Google client class. A default client is created if not given as a parameter. If this is run in an
            environment that has not authenticated via a service account or the likes of gcloud auth login, an
            anonymous client is created for public bucket access only. If valid credentials exist in a session
            they will be used to create the client object.
        """
        super().__init__()
        self.bucket_name = bucket_name
        if client:
            self.client = client
        else:
            try:
                # When a default client can be created, use its credentials. E.g. service accounts running on cloud run,
                # local development/testing.
                logger.info('Creating google bucket client using existing credentials.')
                self.client = Client()
            except (OSError, DefaultCredentialsError) as err:
                # When a default client can't be created, assuming service is being hosted by someone else and google
                # bucket access is via the anonymous client only (i.e. only support public buckets).
                logger.info('Failed to create authenticated google client instance: %s.', str(err))
                logger.info('Creating anonymous google client instance.')
                self.client = Client.create_anonymous_client()

    def _get_json_from_bucket(self: GoogleBucketDataSource, path: str) -> dict | list:
        blob = self.client.get_bucket(self.bucket_name).blob(path)
        return json.loads(blob.download_as_string())

    def send_file_from_bucket(self: GoogleBucketDataSource, file_path: str) -> Response:
        """Fetch any generic blob from a google bucket.

        Parameters
        ----------
        file_path : str
            The "file path" to the google bucket blob.

        Returns
        -------
        Response
            A generic blob object.

        Raises
        ------
        NotFound
            If resource not found.
        """
        bucket = self.client.get_bucket(self.bucket_name)

        blob = bucket.blob(file_path)
        with tempfile.NamedTemporaryFile() as temp:
            try:
                blob.download_to_filename(temp.name)
                return send_file(temp.name, download_name=file_path)
            except GoogleNotFound as e:
                raise NotFound(e)

    def _list_directories(self: GoogleBucketDataSource, root_directory: str) -> List[str]:
        """Return a list of blob directory "paths" that exist under a specific root path.

        Parameters
        ----------
        root_directory : str
            Path prefix - i.e. the first part of the path to search.
            This should not start with a slash. e.g. "api-v1/data-sets/"

        Returns
        -------
        List[str]
            A list of blob paths.
        """
        blobs = self.client.list_blobs(self.bucket_name, prefix=root_directory, delimiter="/")
        # Blobs must be iterated through for the prefixes to be populated.
        # Implementation feature of HTTPIterator
        for _ in blobs:
            pass

        return_list = []
        for root_directory in blobs.prefixes:
            return_list.append(root_directory)

        return return_list

    def file_urls_in_directory(self: GoogleBucketDataSource, path: str) -> Response:
        """List all files in a google bucket "directory" and return as a json response.

        Parameters
        ----------
        path : str
            The directory under which to look for files.

        Returns
        -------
        Response
            A json list with file names (without extensions).
        """
        blobs = self.client.list_blobs(self.bucket_name, prefix=path)

        return_list = []
        for blob in blobs:
            data_set_name = _get_data_set_name_from_path(blob.name)
            # When file_name is None, this indicates a blob that is actually a directory.
            if data_set_name:
                return_list.append(
                    {
                        'name': data_set_name,
                    })

        return jsonify(return_list)

    def _data_sets(self: GoogleBucketDataSource) -> List[dict]:
        """
        Return a list of datasets.

        Returns
        -------
        List[dict]
            A list of datasets with basic meta data. See the abstract implementation for schema reference.

        Raises
        ------
        NotFound
            If individual configuration resource not found. This implies there are some corrupt/incomplete data sets.
        """
        data_sets = []

        data_set_paths = self._list_directories(self.data_set_root_path)

        for data_set_path in data_set_paths:
            # _list_directories always returns data set directories with a trailing '/'
            data_set_name = data_set_path.split('/')[-2]

            try:
                data_set_config = self._get_json_from_bucket(data_set_path + 'configuration.json')
            except GoogleNotFound as e:
                raise NotFound(f"Configuration file is likely missing for data set: {data_set_name}") from e

            data_set_info = {
                'id': data_set_name,
                'name': data_set_name,
                'ui-path': data_set_config['data-set']['ui-path'],
                'ui-type': data_set_config['data-set']['ui-type'],
                'last-modified': data_set_config['data-set']['last-modified']
            }

            data_sets.append(data_set_info)

        return data_sets

    def _data_set_configuration(self: GoogleBucketDataSource, data_set_id: str) -> dict:
        """
        Return data set meta data for a specific data set id.

        Parameters
        ----------
        data_set_id : str
            Unique identifier of the data set

        Returns
        -------
        dict
            A meta data dictionary.
        """
        file_path = self.configuration_template.format(data_set_id=data_set_id)

        return self._get_json_from_bucket(file_path)

    def data_set_tiles(self: GoogleBucketDataSource, tile_definition: Tile) -> Response:
        """
        Get single tile of data.

        Parameters
        ----------
        tile_definition : Tile
            Specification of the tile to retrieve.

        Returns
        -------
        Response
            Http response with object.
        """
        tile_path_template = \
            "api-v1/data-sets/{dataset}/tiles/{resolution}/{x}_{y}{z}_{date}.{extension}"

        tile_path = tile_path_template.format(dataset=tile_definition.data_set_id,
                                              resolution=tile_definition.resolution,
                                              x=tile_definition.x,
                                              y=tile_definition.y,
                                              # Optional string item
                                              z=f"_{tile_definition.z}" if tile_definition.z is not None else "",
                                              date=tile_definition.time,
                                              extension=tile_definition.extension
                                              )

        return self.send_file_from_bucket(tile_path)

    def __eq__(self: GoogleBucketDataSource, other_object: object) -> bool:
        """
        Check object for equality.

        We consider two GoogleBucketSource objects equal if they point to the same bucket.

        Parameters
        ----------
        other_object : object
            The object to compare against this one.

        Returns
        -------
        bool
            Whether the two are equal.
        """
        if isinstance(other_object, GoogleBucketDataSource):
            return other_object.bucket_name == self.bucket_name
        return False



"""Definition of endpoints for v1 of the API."""
from http import HTTPStatus
import json
from multiprocessing.pool import ThreadPool
from typing import Tuple

from flask import jsonify
from flask import request
from flask import Response
from schema import Schema
from schema import SchemaError
from werkzeug.exceptions import InternalServerError

from gateway_4d_viewer import app
from gateway_4d_viewer.api_schema import REGISTER_DATA_SOURCE_SCHEMA
from gateway_4d_viewer.google_bucket_source import GoogleBucketDataSource
from gateway_4d_viewer.interfaces import DataSourceInterface
from gateway_4d_viewer.interfaces import Tile
from gateway_4d_viewer.xcube_server_source import XCubeDataSource

DEFAULT_BUCKET_SOURCE = GoogleBucketDataSource(app.config['DEFAULT_GOOGLE_BUCKET_NAME'])
DATA_SOURCE_DELIMITER = '~'

DATA_SOURCES = {
    "DTA": DEFAULT_BUCKET_SOURCE,
    # Further sources can be pre-registered here such as:
    "ESDL": XCubeDataSource(server_url='https://api.earthsystemdatalab.net/api/4d_viewer'),
    "XCUBE-DEV": XCubeDataSource(server_url='https://deep-esdl.api.dev.brockmann-consult.de/api/4d_viewer')
}


@app.errorhandler(InternalServerError)
def handle_internal_server(error: InternalServerError) -> Response:
    """
    Handle InternalServerError to pass error details to client.

    Parameters
    ----------
    error : InternalServerError
        The internally raised error

    Returns
    -------
    Response
        InternalServerError response with additional error information.
    """
    response = jsonify({"Internal error": error.description})
    response.status_code = error.code
    return response


def _split_api_id(compound_data_set_id: str) -> Tuple[DataSourceInterface, str]:
    """Return data source instance and data set name.

    For now this API encodes the data source into a compound data set id. We simply split the string based on the
    delimiter to deduce the dataset str and to find and return the correct data source instance.

    Parameters
    ----------
    compound_data_set_id : str
        A compound data-source + data-set string identifier

    Returns
    -------
    tuple[DataSourceInterface, str]
        (Instance of the DataSourceInterface interface, name of dataset)
    """
    key_split = compound_data_set_id.split(DATA_SOURCE_DELIMITER)
    data_source = DATA_SOURCES[key_split[0]]
    data_set = DATA_SOURCE_DELIMITER.join(key_split[1:])
    return data_source, data_set


@app.route("/api-v1/data-sets", methods=["GET"])
def data_sets() -> Response:
    """
    Get a list of available datasets.

    Returns
    -------
    Response
        A collection of meta data for each data set.
    """
    data_sets = []

    # Determine each data-source's list of datasets in parallel. Each call can be slow, but they're almost entirely
    # bound by processes outside of this service. We may seek to increase the number of threads if we add more
    # data sources.
    pool = ThreadPool(processes=4)
    results = pool.map(lambda data_source: DATA_SOURCES[data_source].data_sets(), DATA_SOURCES)

    for data_source, result in zip(DATA_SOURCES, results):
        # Override id field to use a compound source + data set key. This simplifies future api calls and makes
        # deducing data source from request calls easy in those calls.
        for data_set in result:
            data_set['id'] = data_source + DATA_SOURCE_DELIMITER + data_set['id']

        data_sets += result
        response = jsonify(data_sets)

    return response


@app.route("/api-v1/data-sets/<compound_data_set_id>/configuration", methods=["GET"])
def data_set_configuration(compound_data_set_id: str) -> Response:
    """
    Return dataset configuration.

    Parameters
    ----------
    compound_data_set_id : str
        A compound data-source + data-set string identifier

    Returns
    -------
    Response
        Json response with data set meta data.
    """
    data_source, data_set = _split_api_id(compound_data_set_id)

    return data_source.data_set_configuration(data_set)


@app.route("/api-v1/data-sets/<compound_data_set_id>/tiles/<resolution>_<time>_<x>_<y>_<extension>",
           defaults={'z': None})
@app.route("/api-v1/data-sets/<compound_data_set_id>/tiles/<resolution>_<time>_<x>_<y>_<z>_<extension>",
           methods=["GET"])
def data_set_tiles(compound_data_set_id: str,   # noqa:CFQ002 - many arguments are necessary for endpoint.
                   resolution: str, time: str, x: str, y: str, z: str, extension: str) -> Response:
    """Return an individual tile.

    Parameters
    ----------
    compound_data_set_id : str
        A compound data-source + data-set string identifier
    resolution : str
        The desired resolution, defined as '0', '1', '2' etc as defined by the tile grid.
    time : str
        The time index into the list of times in the tile grid meta data ('0', '1', '2').
    x : str
        The x index as defined by the tile grid ('0', '1', '2').
    y : str
        The y index as defined by the tile grid ('0', '1', '2').
    z : str
        The z index (vertical) as defined by the tile grid ('0', '1', '2'). From an API perspective, z is optional.
    extension : str
        File type extension (typically png, or b)
    Returns
    -------
    Response
        A single tile. The exact object format will differ based on file type.
    """
    data_source, data_set = _split_api_id(compound_data_set_id)
    tile_definition = Tile(data_set_id=data_set, resolution=resolution,
                           x=x, y=y, z=z, extension=extension, time=time)

    return data_source.data_set_tiles(tile_definition)


@app.route("/api-v1/color-maps", methods=["GET"])
def color_maps() -> Response:
    """
    Return list of all available color maps.

    As the url is returned to the user, no need to implement a separate
    GET route for the file itself (authentication pending).

    Returns
    -------
    Response
        Response json containing [ {“name”: “”, “link”: “”},{…} ]
    """
    return DEFAULT_BUCKET_SOURCE.file_urls_in_directory("api-v1/colour-maps/")


@app.route("/api-v1/color-maps/<color_map_id>", methods=["GET"])
def color_map(color_map_id: str) -> Response:
    """
    Return individual colour map image.

    Parameters
    ----------
    color_map_id : str
        The color map id to be returned.

    Returns
    -------
    Response
        image type
    """
    return DEFAULT_BUCKET_SOURCE.send_file_from_bucket(f"api-v1/colour-maps/{color_map_id}.png")


@app.route("/api-v1/configurations", methods=["GET"])
def configurations() -> Response:
    """
    Return list of available configurations.

    Returns
    -------
    Response
        Response json containing [ {“name”: “”, “link”: “”},{…} ]
    """
    return DEFAULT_BUCKET_SOURCE.file_urls_in_directory("api-v1/configurations/")


@app.route("/api-v1/configurations/<configuration_id>", methods=["GET"])
def configuration(configuration_id: str) -> Response:
    """
    Get configuration json used for UI world and story setup.

    Parameters
    ----------
    configuration_id : str
        Configuration Id.

    Returns
    -------
    Response
        json containing ui config.
    """
    return DEFAULT_BUCKET_SOURCE.send_file_from_bucket(f"api-v1/configurations/{configuration_id}.json")


@app.route("/register-data-source/<source_id>", methods=["POST"])
def register_data_source(source_id: str) -> Response:
    """
    Register new data source.

    We can register either a google bucket source, by specifying the name of the bucket, or an xcube server source,
    by specifying the server's externally accessible URL.

    Parameters
    ----------
    source_id : str
        Unique identifier of the new source.

    Returns
    -------
    Response
        Failure message - if the request format is invalid or if a source with the same ID but different content (i.e.
        different bucket name of xcube server url) is already present in list of sources.
        Success message otherwise.
    """
    request_data = json.loads(request.data)
    try:
        Schema(REGISTER_DATA_SOURCE_SCHEMA).validate(request_data)
    except SchemaError as err:
        return f'Format invalid: Error: {err}', HTTPStatus.BAD_REQUEST

    data_source_type = request_data["data_source_type"]
    if data_source_type == "google_bucket_data_source":
        new_data_source = GoogleBucketDataSource(bucket_name=request_data["bucket_name"])
    elif data_source_type == "xcube_server_data_source":
        new_data_source = XCubeDataSource(server_url=request_data["server_url"])

    if source_id in DATA_SOURCES and DATA_SOURCES[source_id] != new_data_source:
        return f"Data source {source_id} already present.", HTTPStatus.BAD_REQUEST

    DATA_SOURCES[source_id] = new_data_source
    return f"Data source {source_id} added.", HTTPStatus.OK


@app.route("/deregister-data-source/<source_id>", methods=["DELETE"])
def deregister_data_source(source_id: str) -> Response:
    """
    Delete data source from list of sources.

    Parameters
    ----------
    source_id : str
        ID of source to delete. Same as the source key in the DATA_SOURCES dictionary.

    Returns
    -------
    Response
        Success/failure message.
    """
    if source_id in DATA_SOURCES:
        DATA_SOURCES.pop(source_id)
        return f"Data source {source_id} deleted.", HTTPStatus.OK
    else:
        return f"Data source {source_id} not present, cannot delete.", HTTPStatus.BAD_REQUEST



"""Unit tests for gateway_4d_viewer.helpers.datasets.py."""
from unittest.mock import patch

from gateway_4d_viewer import dataset_helpers
from gateway_4d_viewer.google_bucket_source import GoogleBucketDataSource


def test_get_datasets():
    with patch.object(GoogleBucketDataSource, '_list_directories') as mock_list_blobs_with_prefix:
        mock_list_blobs_with_prefix.return_value = ['gateway-4d-viewer/datasets/A/', 'gateway-4d-viewer/datasets/B/']
        result = dataset_helpers.get_datasets(None, "not_used", 'gateway-4d-viewer/datasets/')

        assert result == '{"datasets": [{"id": "A", "title": "A"}, {"id": "B", "title": "B"}]}'



"""Setup script."""
import os

from setuptools import find_packages
from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

if os.path.exists("full_version.txt"):
    with open("full_version.txt", "r", encoding="utf-8") as fh:
        """
        Note that this file is generated by the CI chain based on the git tag
        (by .github/scripts/define_new_version_number.py)
        It should not be present in the repository by default.
        """
        version_number = fh.read()
else:
    version_number = 'v0.0.0'  # default value when under development

setup(
    name='gateway_4d_viewer',
    version=version_number,
    description="A flask-based gateway for serving data to Earthwave's 4d viewer client.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Earthwave Ltd',
    author_email='info@earthwave.co.uk',
    url='https://earthwave.co.uk/',
    python_requires=">=3.9",
    packages=find_packages(),
    # note requirements listed ininstall_requires should be the *minimum required*
    # in order to allow pip to resolve multiple installed packages properly.
    # requirements.txt should contain a specific known working version instead.
    install_requires=[
        'flask',
        'flask_cors',
        'gunicorn',
        'google-cloud-storage',
        'google-api-core',
        'schema'
    ],
)



"""Simple entry point for application."""
import logging

from gateway_4d_viewer import app

if __name__ == "__main__":
    app.run(debug=True)
else:
    # setup logging for when gunicorn is running the app
    gunicorn_logger = logging.getLogger('gunicorn.error')
    app.logger.handlers = gunicorn_logger.handlers
    app.logger.setLevel(gunicorn_logger.level)



"""Basic application initialisation."""
from flask import Flask
from flask_cors import CORS


app = Flask(__name__)
app.config.from_pyfile('config.py')
CORS(app, resources={r"/*": {"origins": "*"}})

# The import ordering is on purpose to avoid circular imports and simplify the project structure.
# c.f. https://flask.palletsprojects.com/en/latest/patterns/packages/
from gateway_4d_viewer import routes_v1  # noqa: F401, E402
